@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2025 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
+
 #include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
 
@@ -7,6 +29,9 @@
 
 #define ALLOWED_TEXT_POS_ERROR (0.001f)
 
+#define ENC_IDENTITY 0
+#define ENC_UNICODE 1
+
 typedef struct pdf_device pdf_device;
 
 typedef struct
@@ -78,6 +103,7 @@ struct pdf_device
 	int num_cid_fonts;
 	int max_cid_fonts;
 	fz_font **cid_fonts;
+	int *cid_fonts_enc;
 
 	int num_alphas;
 	int max_alphas;
@@ -190,6 +216,13 @@ static void
 pdf_dev_path(fz_context *ctx, pdf_device *pdev, const fz_path *path)
 {
 	gstate *gs = CURRENT_GSTATE(pdev);
+	fz_rect bounds;
+
+	if (fz_path_is_rect_with_bounds(ctx, path, fz_identity, &bounds))
+	{
+		fz_append_printf(ctx, gs->buf, "%g %g %g %g re\n", bounds.x0, bounds.y0, bounds.x1-bounds.x0, bounds.y1-bounds.y0);
+		return;
+	}
 
 	fz_walk_path(ctx, path, &pdf_dev_path_proc, (void *)gs->buf);
 }
@@ -283,6 +316,8 @@ pdf_dev_alpha(fz_context *ctx, pdf_device *pdev, float alpha, int stroke)
 	if (gs->alpha[stroke] == alpha)
 		return;
 
+	gs->alpha[stroke] = alpha;
+
 	/* Have we sent such an alpha before? */
 	for (i = 0; i < pdev->num_alphas; i++)
 		if (pdev->alphas[i].alpha == alpha && pdev->alphas[i].stroke == stroke)
@@ -327,20 +362,23 @@ pdf_dev_alpha(fz_context *ctx, pdf_device *pdev, float alpha, int stroke)
 }
 
 static int
-pdf_dev_add_font_res(fz_context *ctx, pdf_device *pdev, fz_font *font)
+pdf_dev_find_font_res(fz_context *ctx, pdf_device *pdev, fz_font *font)
 {
-	pdf_obj *fres;
-	char text[32];
 	int k;
-	int num;
 
 	/* Check if we already had this one */
 	for (k = 0; k < pdev->num_cid_fonts; k++)
 		if (pdev->cid_fonts[k] == font)
 			return k;
 
-	/* This will add it to the xref if needed */
-	fres = pdf_add_cid_font(ctx, pdev->doc, font);
+	return -1;
+}
+
+static int
+pdf_dev_add_font_res_imp(fz_context *ctx, pdf_device *pdev, fz_font *font, pdf_obj *fres, int enc)
+{
+	char text[32];
+	int num;
 
 	/* Not there so add to resources */
 	fz_snprintf(text, sizeof(text), "Font/F%d", pdev->num_cid_fonts);
@@ -353,13 +391,54 @@ pdf_dev_add_font_res(fz_context *ctx, pdf_device *pdev, fz_font *font)
 		if (newmax == 0)
 			newmax = 4;
 		pdev->cid_fonts = fz_realloc_array(ctx, pdev->cid_fonts, newmax, fz_font*);
+		pdev->cid_fonts_enc = fz_realloc_array(ctx, pdev->cid_fonts_enc, newmax, int);
 		pdev->max_cid_fonts = newmax;
 	}
 	num = pdev->num_cid_fonts++;
 	pdev->cid_fonts[num] = fz_keep_font(ctx, font);
+	pdev->cid_fonts_enc[num] = enc;
 	return num;
 }
 
+static int
+pdf_dev_add_substitute_font_res(fz_context *ctx, pdf_device *pdev, fz_font *font)
+{
+	pdf_obj *fres;
+	int k;
+
+	/* Check if we already had this one */
+	k = pdf_dev_find_font_res(ctx, pdev, font);
+	if (k >= 0)
+		return k;
+
+	/* This will add it to the xref if needed */
+	if (font->flags.cjk)
+		fres = pdf_add_cjk_font(ctx, pdev->doc, font, font->flags.cjk_lang, 0, font->flags.is_serif);
+	else
+		fres = pdf_add_substitute_font(ctx, pdev->doc, font);
+
+	/* And add to the resource dictionary. */
+	return pdf_dev_add_font_res_imp(ctx, pdev, font, fres, ENC_UNICODE);
+}
+
+static int
+pdf_dev_add_embedded_font_res(fz_context *ctx, pdf_device *pdev, fz_font *font)
+{
+	pdf_obj *fres;
+	int k;
+
+	/* Check if we already had this one */
+	k = pdf_dev_find_font_res(ctx, pdev, font);
+	if (k >= 0)
+		return k;
+
+	/* This will add it to the xref if needed */
+	fres = pdf_add_cid_font(ctx, pdev->doc, font);
+
+	/* And add to the resource dictionary. */
+	return pdf_dev_add_font_res_imp(ctx, pdev, font, fres, ENC_IDENTITY);
+}
+
 static void
 pdf_dev_font(fz_context *ctx, pdf_device *pdev, fz_font *font, fz_matrix trm)
 {
@@ -370,14 +449,16 @@ pdf_dev_font(fz_context *ctx, pdf_device *pdev, fz_font *font, fz_matrix trm)
 	if (gs->font >= 0 && pdev->cid_fonts[gs->font] == font && gs->font_size == font_size)
 		return;
 
+	// TODO: vertical wmode
+
 	if (fz_font_t3_procs(ctx, font))
-		fz_throw(ctx, FZ_ERROR_GENERIC, "pdf device does not support type 3 fonts");
-	if (fz_font_flags(font)->ft_substitute)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "pdf device does not support substitute fonts");
-	if (!pdf_font_writing_supported(font))
-		fz_throw(ctx, FZ_ERROR_GENERIC, "pdf device does not support font types found in this file");
+		fz_throw(ctx, FZ_ERROR_UNSUPPORTED, "pdf device does not support type 3 fonts");
+
+	if (fz_font_flags(font)->ft_substitute || !pdf_font_writing_supported(ctx, font))
+		gs->font = pdf_dev_add_substitute_font_res(ctx, pdev, font);
+	else
+		gs->font = pdf_dev_add_embedded_font_res(ctx, pdev, font);
 
-	gs->font = pdf_dev_add_font_res(ctx, pdev, font);
 	gs->font_size = font_size;
 
 	fz_append_printf(ctx, gs->buf, "/F%d %g Tf\n", gs->font, gs->font_size);
@@ -433,6 +514,7 @@ pdf_dev_text_span(fz_context *ctx, pdf_device *pdev, fz_text_span *span)
 	fz_matrix inv_tfs;
 	fz_point d;
 	float adv;
+	int enc;
 	int dx, dy;
 	int i;
 
@@ -451,12 +533,16 @@ pdf_dev_text_span(fz_context *ctx, pdf_device *pdev, fz_text_span *span)
 	inv_tm = fz_invert_matrix(tm);
 	inv_trm = fz_invert_matrix(trm);
 
+	enc = pdev->cid_fonts_enc[gs->font];
+
 	fz_append_printf(ctx, gs->buf, "%M Tm\n[<", &tm);
 
 	for (i = 0; i < span->len; ++i)
 	{
 		fz_text_item *it = &span->items[i];
-		if (it->gid < 0)
+		if (enc == ENC_IDENTITY && it->gid < 0)
+			continue;
+		if (enc == ENC_UNICODE && it->ucs < 0)
 			continue;
 
 		/* transform difference from expected pen position into font units. */
@@ -489,14 +575,19 @@ pdf_dev_text_span(fz_context *ctx, pdf_device *pdev, fz_text_span *span)
 
 		if (fz_font_t3_procs(ctx, span->font))
 			fz_append_printf(ctx, gs->buf, "%02x", it->gid);
-		else
+		else if (enc == ENC_IDENTITY)
 			fz_append_printf(ctx, gs->buf, "%04x", it->gid);
+		else if (enc == ENC_UNICODE)
+			fz_append_printf(ctx, gs->buf, "%04x", it->ucs);
 
-		adv = fz_advance_glyph(ctx, span->font, it->gid, span->wmode);
-		if (span->wmode == 0)
-			trm = fz_pre_translate(trm, adv, 0);
-		else
-			trm = fz_pre_translate(trm, 0, adv);
+		if (it->gid != -1)
+		{
+			adv = fz_advance_glyph(ctx, span->font, it->gid, span->wmode);
+			if (span->wmode == 0)
+				trm = fz_pre_translate(trm, adv, 0);
+			else
+				trm = fz_pre_translate(trm, 0, adv);
+		}
 	}
 
 	fz_append_string(ctx, gs->buf, ">]TJ\n");
@@ -733,6 +824,7 @@ pdf_dev_stroke_text(fz_context *ctx, fz_device *dev, const fz_text *text, const
 	pdf_dev_ctm(ctx, pdev, ctm);
 	pdf_dev_alpha(ctx, pdev, alpha, 1);
 	pdf_dev_color(ctx, pdev, colorspace, color, 1, color_params);
+	pdf_dev_stroke_state(ctx, pdev, stroke);
 
 	for (span = text->head; span; span = span->next)
 	{
@@ -867,6 +959,8 @@ pdf_dev_fill_shade(fz_context *ctx, fz_device *dev, fz_shade *shade, fz_matrix c
 {
 	pdf_device *pdev = (pdf_device*)dev;
 
+	fz_warn(ctx, "the pdf device does not support shadings; output may be incomplete");
+
 	/* FIXME */
 	pdf_dev_end_text(ctx, pdev);
 }
@@ -913,6 +1007,8 @@ pdf_dev_clip_image_mask(fz_context *ctx, fz_device *dev, fz_image *image, fz_mat
 {
 	pdf_device *pdev = (pdf_device*)dev;
 
+	fz_warn(ctx, "the pdf device does not support image masks; output may be incomplete");
+
 	/* FIXME */
 	pdf_dev_end_text(ctx, pdev);
 	pdf_dev_push(ctx, pdev);
@@ -947,6 +1043,8 @@ pdf_dev_begin_mask(fz_context *ctx, fz_device *dev, fz_rect bbox, int luminosity
 
 	pdf_dev_end_text(ctx, pdev);
 
+	pdf_dev_ctm(ctx, pdev, fz_identity);
+
 	/* Make a new form to contain the contents of the softmask */
 	pdf_dev_new_form(ctx, &form_ref, pdev, bbox, 0, 0, 1, colorspace);
 
@@ -985,13 +1083,16 @@ pdf_dev_begin_mask(fz_context *ctx, fz_device *dev, fz_rect bbox, int luminosity
 }
 
 static void
-pdf_dev_end_mask(fz_context *ctx, fz_device *dev)
+pdf_dev_end_mask(fz_context *ctx, fz_device *dev, fz_function *tr)
 {
 	pdf_device *pdev = (pdf_device*)dev;
 	pdf_document *doc = pdev->doc;
 	gstate *gs = CURRENT_GSTATE(pdev);
 	pdf_obj *form_ref = (pdf_obj *)gs->on_pop_arg;
 
+	if (tr)
+		fz_warn(ctx, "Ignoring Transfer function");
+
 	/* Here we do part of the pop, but not all of it. */
 	pdf_dev_end_text(ctx, pdev);
 	fz_append_string(ctx, gs->buf, "Q\n");
@@ -1007,28 +1108,31 @@ static void
 pdf_dev_begin_group(fz_context *ctx, fz_device *dev, fz_rect bbox, fz_colorspace *cs, int isolated, int knockout, int blendmode, float alpha)
 {
 	pdf_device *pdev = (pdf_device*)dev;
-	pdf_document *doc = pdev->doc;
 	int num;
 	pdf_obj *form_ref;
 	gstate *gs;
 
 	pdf_dev_end_text(ctx, pdev);
 
+	pdf_dev_ctm(ctx, pdev, fz_identity);
+
 	num = pdf_dev_new_form(ctx, &form_ref, pdev, bbox, isolated, knockout, alpha, cs);
 
 	/* Do we have an appropriate blending extgstate already? */
 	{
 		char text[32];
 		pdf_obj *obj;
-		fz_snprintf(text, sizeof(text), "ExtGState/BlendMode%d", blendmode);
-		obj = pdf_dict_getp(ctx, pdev->resources, text);
+		pdf_obj *egs = pdf_dict_get(ctx, pdev->resources, PDF_NAME(ExtGState));
+		if (egs == NULL)
+			egs = pdf_dict_put_dict(ctx, pdev->resources, PDF_NAME(ExtGState), 4);
+		fz_snprintf(text, sizeof(text), "BlendMode%d", blendmode);
+		obj = pdf_dict_gets(ctx, egs, text);
 		if (obj == NULL)
 		{
 			/* No, better make one */
-			obj = pdf_new_dict(ctx, doc, 2);
+			obj = pdf_dict_puts_dict(ctx, egs, text, 2);
 			pdf_dict_put(ctx, obj, PDF_NAME(Type), PDF_NAME(ExtGState));
 			pdf_dict_put_name(ctx, obj, PDF_NAME(BM), fz_blendmode_name(blendmode));
-			pdf_dict_putp_drop(ctx, pdev->resources, text, obj);
 		}
 	}
 
@@ -1106,13 +1210,14 @@ pdf_dev_drop_device(fz_context *ctx, fz_device *dev)
 
 	pdf_drop_obj(ctx, pdev->resources);
 	fz_free(ctx, pdev->cid_fonts);
+	fz_free(ctx, pdev->cid_fonts_enc);
 	fz_free(ctx, pdev->image_indices);
 	fz_free(ctx, pdev->groups);
 	fz_free(ctx, pdev->alphas);
 	fz_free(ctx, pdev->gstates);
 }
 
-fz_device *pdf_new_pdf_device(fz_context *ctx, pdf_document *doc, fz_matrix topctm, fz_rect mediabox, pdf_obj *resources, fz_buffer *buf)
+fz_device *pdf_new_pdf_device(fz_context *ctx, pdf_document *doc, fz_matrix topctm, pdf_obj *resources, fz_buffer *buf)
 {
 	pdf_device *dev = fz_new_derived_device(ctx, pdf_device);
 
@@ -1149,19 +1254,18 @@ fz_device *pdf_new_pdf_device(fz_context *ctx, pdf_document *doc, fz_matrix topc
 
 	fz_try(ctx)
 	{
-		if (buf)
-			buf = fz_keep_buffer(ctx, buf);
-		else
-			buf = fz_new_buffer(ctx, 256);
 		dev->doc = doc;
 		dev->resources = pdf_keep_obj(ctx, resources);
 		dev->gstates = fz_malloc_struct(ctx, gstate);
-		dev->gstates[0].buf = buf;
+		if (buf)
+			dev->gstates[0].buf = fz_keep_buffer(ctx, buf);
+		else
+			dev->gstates[0].buf = fz_new_buffer(ctx, 256);
 		dev->gstates[0].ctm = fz_identity; // XXX
 		dev->gstates[0].colorspace[0] = fz_device_gray(ctx);
 		dev->gstates[0].colorspace[1] = fz_device_gray(ctx);
-		dev->gstates[0].color[0][0] = 1;
-		dev->gstates[0].color[1][0] = 1;
+		dev->gstates[0].color[0][0] = 0;
+		dev->gstates[0].color[1][0] = 0;
 		dev->gstates[0].alpha[0] = 1.0f;
 		dev->gstates[0].alpha[1] = 1.0f;
 		dev->gstates[0].font = -1;
@@ -1169,12 +1273,11 @@ fz_device *pdf_new_pdf_device(fz_context *ctx, pdf_document *doc, fz_matrix topc
 		dev->max_gstates = 1;
 
 		if (!fz_is_identity(topctm))
-			fz_append_printf(ctx, buf, "%M cm\n", &topctm);
+			fz_append_printf(ctx, dev->gstates[0].buf, "%M cm\n", &topctm);
 	}
 	fz_catch(ctx)
 	{
-		fz_drop_buffer(ctx, buf);
-		fz_free(ctx, dev);
+		fz_drop_device(ctx, &dev->super);
 		fz_rethrow(ctx);
 	}
 
@@ -1184,7 +1287,9 @@ fz_device *pdf_new_pdf_device(fz_context *ctx, pdf_document *doc, fz_matrix topc
 fz_device *pdf_page_write(fz_context *ctx, pdf_document *doc, fz_rect mediabox, pdf_obj **presources, fz_buffer **pcontents)
 {
 	fz_matrix pagectm = { 1, 0, 0, -1, -mediabox.x0, mediabox.y1 };
-	*presources = pdf_new_dict(ctx, doc, 0);
-	*pcontents = fz_new_buffer(ctx, 0);
-	return pdf_new_pdf_device(ctx, doc, pagectm, mediabox, *presources, *pcontents);
+	if (!*presources)
+		*presources = pdf_new_dict(ctx, doc, 0);
+	if (!*pcontents)
+		*pcontents = fz_new_buffer(ctx, 0);
+	return pdf_new_pdf_device(ctx, doc, pagectm, *presources, *pcontents);
 }
